package main

import (
	"context"
	"fmt"
	"git.sch9.ru/new_gate/ms-auth/config"
	"git.sch9.ru/new_gate/ms-auth/internal/models"
	"git.sch9.ru/new_gate/ms-auth/internal/users/delivery/rest"
	usersRepository "git.sch9.ru/new_gate/ms-auth/internal/users/repository"
	usersUseCase "git.sch9.ru/new_gate/ms-auth/internal/users/usecase"
	"git.sch9.ru/new_gate/ms-auth/pkg"
	userv1 "git.sch9.ru/new_gate/ms-auth/proto/user/v1"
	"github.com/gofiber/fiber/v2"
	fiberlogger "github.com/gofiber/fiber/v2/middleware/logger"
	"github.com/ilyakaznacheev/cleanenv"
	_ "github.com/jackc/pgx/v5/stdlib"
	"go.uber.org/zap"
	"os"
	"os/signal"
	"syscall"
)

// InterceptorLogger adapts zap logger to interceptor logger.
// This code is simple enough to be copied and not imported.
//func InterceptorLogger(l *zap.Logger) fiber.Handler {
//	return func(ctx context.Context, lvl logging.Level, msg string, fields ...any) {
//		f := make([]zap.Field, 0, len(fields)/2)
//
//		for i := 0; i < len(fields); i += 2 {
//			key := fields[i]
//			value := fields[i+1]
//
//			switch v := value.(type) {
//			case string:
//				f = append(f, zap.String(key.(string), v))
//			case int:
//				f = append(f, zap.Int(key.(string), v))
//			case bool:
//				f = append(f, zap.Bool(key.(string), v))
//			default:
//				f = append(f, zap.Any(key.(string), v))
//			}
//		}
//
//		logger := l.WithOptions(zap.AddCallerSkip(1)).With(f...)
//
//		switch lvl {
//		case logging.LevelDebug:
//			logger.Debug(msg)
//		case logging.LevelInfo:
//			logger.Info(msg)
//		case logging.LevelWarn:
//			logger.Warn(msg)
//		case logging.LevelError:
//			logger.Error(msg)
//		default:
//			panic(fmt.Sprintf("unknown level %v", lvl))
//		}
//	})
//}

func main() {
	var cfg config.Config
	err := cleanenv.ReadConfig(".env", &cfg)
	if err != nil {
		panic(fmt.Sprintf("error reading config: %s", err.Error()))
	}

	var logger *zap.Logger
	if cfg.Env == "prod" {
		logger = zap.Must(zap.NewProduction())
	} else if cfg.Env == "dev" {
		logger = zap.Must(zap.NewDevelopment())
	} else {
		panic(fmt.Sprintf(`error reading config: env expected "prod" or "dev", got "%s"`, cfg.Env))
	}

	logger.Info("connecting to postgres")
	db, err := pkg.NewPostgresDB(cfg.PostgresDSN)
	if err != nil {
		logger.Fatal(fmt.Sprintf("error connecting to postgres: %s", err.Error()))
	}
	defer db.Close()
	logger.Info("successfully connected to postgres")

	logger.Info("connecting to redis")
	vk, err := pkg.NewValkeyClient(cfg.RedisDSN)
	if err != nil {
		logger.Fatal(fmt.Sprintf("error connecting to redis: %s", err.Error()))
	}
	logger.Info("successfully connected to redis")

	userRepo := usersRepository.NewUserRepository(db)

	_, err = userRepo.C().CreateUser(context.Background(), cfg.AdminUsername, cfg.AdminPassword, models.RoleAdmin)
	if err != nil {
		logger.Error(fmt.Sprintf("error creating admin user: %s", err.Error()))
	}

	sessionRepo := usersRepository.NewValkeyRepository(vk)
	userUC := usersUseCase.NewUseCase(userRepo, sessionRepo, cfg)

	server := fiber.New()

	userv1.RegisterHandlersWithOptions(server, rest.NewUserHandlers(userUC, cfg.JWTSecret), userv1.FiberServerOptions{
		Middlewares: []userv1.MiddlewareFunc{
			fiberlogger.New(),
			rest.AuthMiddleware(cfg.JWTSecret, userUC),
			//cors.New(cors.Config{
			//	AllowOrigins:     "http://localhost:3000",
			//	AllowMethods:     "GET,POST,PUT,DELETE,OPTIONS",
			//	AllowHeaders:     "Content-Type,Set-Cookie,Credentials",
			//	AllowCredentials: true,
			//}),
		},
	})

	go func() {
		err := server.Listen(cfg.Address)
		if err != nil {
			logger.Fatal(fmt.Sprintf("error starting server: %s", err.Error()))
		}
	}()

	logger.Info(fmt.Sprintf("server started on %s", cfg.Address))

	stop := make(chan os.Signal, 1)
	signal.Notify(stop, syscall.SIGTERM, syscall.SIGINT)

	<-stop
}
